// A string-to-unsigned dictionary represented as a binary search tree.
#pragma once

#include <memory>
#include <stdexcept>
#include <string>
#include <vector>

namespace dictionary {

class Dict {
public:
    // Default constructor:
    Dict();

    // Copy constructor:
    Dict(const Dict&);

    // Copy-assignment operator:
    Dict& operator=(const Dict&);

    // Frees the tree's memory when we're done with it:
    ~Dict();

    // Swaps two dictionaries in O(1) time.
    void swap(Dict&);

    bool empty() const;
    size_t size() const;

    // Removes all mappings from the tree
    void clear();

    std::vector<std::string> get_keys() const;
    bool member(const std::string&) const;
    unsigned lookup(const std::string&) const;

    unsigned& lookup(const std::string&);
    void insert(const std::string&, unsigned);

    // Lookup, but inserts 0 if not found.
    unsigned& operator[](const std::string&);

private:
    struct Node;

    // Static functions are functions that don't operate on an instance of
    // the class of which they are members. That is, these functions do
    // not get a Dict when they are called:
    static Node* find_subtree_(Node*, const std::string&);
    static Node* & find_subtree_mut_(Node*&, const std::string&);
    static void get_keys_helper_(Node*, std::vector<std::string>&);

    Node* root_;
    size_t size_;
};

// An exception thrown by Dict::lookup when the key
// isn't found.
struct Not_found : std::runtime_error
{
    Not_found(const std::string& key) :
            runtime_error("key not found: " + key) { }
};

// Overloading swap allows algorithms such as sorting to work efficiently.
void swap(Dict&, Dict&);

}